/* ----------------------------------------------------------------------------
 *
 * Raima Database Manager
 *
 * Copyright (c) 2012 Raima Inc.,  All rights reserved.
 *
 * Use of this software, whether in source code format, or in executable,
 * binary object code form, is governed by the Raima LICENSE which
 * is fully described in the LICENSE.TXT file, included within this
 * distribution of files.
 *
 * ----------------------------------------------------------------------------
 */
#include <stdio.h>
#include <pthread.h>

/* rdm.h is required in all files using the RDM Database */
#include "rdm.h"

/* core17_db.h was generated from the DDL file core17_db.ddl by the ddlp.
 * utility.
 */
#include "core17_db.h"

/* core17_db_dbd.h was generated from the DDL
 * file core17_db.ddl by the ddlp utility.  It is required to use the d_open_ptr API
 */
#include "core17_db_dbd.h"

/* EXAMPLE - Core for C */

/* This example shows multi-user access to a database using separate processes.
 * and a remote tfserver.  The connections will use TCP/IP to communicate with
 * the tfserver to do their work.  After starting the tfserver two writer 
 * processes are created.  The first begins inserting values into
 * the table.  The second waits until the table contains more than 100 rows
 * and then begins deleting the oldest.  A reader process is also created
 * that prints out the total number of records, the max value, and the min
 * value.  Each thread uses regular locking.
 *
 * The DDL (Database Definition Language) is in core17_db.ddl. The database
 * schema is embedded into the application and the database will be created 
 * the first time the application is run.  Subsequent runs will use the now
 * existing database
 *
 * The following RDM Functions are illustrated:
 *      d_tfsinitEx/d_tfsterm
 *      d_opentask/d_closetask
 *      d_open_ptr/d_close
 *      d_initialize
 *      d_reclock/d_recfree
 *      d_trbegin/d_trend/d_trabort
 *      d_fillnew
 *      d_delete
 *      d_rectot
 *      d_keyfrst/d_keynext
 *      d_recread
 *      d_errorname
 *      d_errorstr
 *
 * They are fully documented in http://docs.raima.com/rdm/.
 *
 * Each of these functions returns an integer status code, where the
 * value S_OKAY indicates a successful call.
 *
 * For simplicity, this example does not check all return codes, but good
 * programming practices would dictate that they are checked after each
 * RDM call.
 *
 */

static void Inserter(void)
{
    DB_TASK      *task;
    uint32_t      ii;
    int32_t       rdm_status;
    struct rec01  rec01_rec;
    TFS_HANDLE    hTFS;

    printf("Inserter process\n");

    /* Initialize the application to use the external TFS (RPC) that we 
     * started
     */
    d_tfsinitEx(NULL, NULL, &hTFS, TFS_TYPE_RPC);

    /* Create a task for this thread */
    rdm_status = d_opentask(&task);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Error open DB_TASK in InserterThreadFcn - %s: %s\n", d_errorname(rdm_status), d_errorstr(rdm_status));

        exit(EXIT_FAILURE);
    }

    /* Open the database in shared mode */
    rdm_status = d_open("core17_db", "s", task);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Error opening database in InserterThreadFcn - %s: %s\n", d_errorname(rdm_status), d_errorstr(rdm_status));

        d_closetask(task);
        exit(EXIT_FAILURE);
    }

    /* Loop until we get an error or are told to quit */
    for (ii = 0; rdm_status == S_OKAY; ii++)
    {
        /* Start a transaction */
        rdm_status = d_trbegin("AddRecord", task);
        if (rdm_status == S_OKAY)
        {
            /* Lock the REC01 record */
            rdm_status = d_reclock(REC01, "w", task, CURR_DB);
            if (rdm_status == S_OKAY)
            {
                /* Add a new record instance */
                rec01_rec.field01 = ii;
                rdm_status = d_fillnew(REC01, &rec01_rec, task, CURR_DB);
                
                /* Commit or rollback the tranaction */
                if (rdm_status == S_OKAY)
                    d_trend(task);
                else
                    d_trabort(task);
            }
        }
    }
}

static void Deleter(void)
{
    DB_TASK    *task;
    int32_t     rdm_status;
    uint64_t    rec_count = 0;
    TFS_HANDLE  hTFS;

    printf("Deleter process\n");

    /* Initialize the application to use the external TFS (RPC) that we 
     * started
     */
    d_tfsinitEx(NULL, NULL, &hTFS, TFS_TYPE_RPC);

    /* Create a task for this thread */
    rdm_status = d_opentask(&task);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Error open DB_TASK in DeleterThreadFcn - %s: %s\n", d_errorname(rdm_status), d_errorstr(rdm_status));

        exit(EXIT_FAILURE);
    }

    /* Open the database in shared mode */
    rdm_status = d_open("core17_db", "s", task);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Error opening database in DeleterThreadFcn - %s: %s\n", d_errorname(rdm_status), d_errorstr(rdm_status));

        d_closetask(task);
        exit(EXIT_FAILURE);
    }

    /* Loop until we get an error or are told to quit */
    while (rdm_status == S_OKAY)
    {
        /* Check the total number of REC01 record instances */
        rdm_status = d_trrobegin(task, CURR_DB, NULL);
        if(rdm_status == S_OKAY)
        {
            rdm_status = d_rectot(REC01, &rec_count, task, CURR_DB);
            d_trroend(task, CURR_DB);
        }

        /* When we have more than 100 records start deleting the record with
         * the smallest FIELD01 value
         */
        if( rdm_status == S_OKAY && rec_count > 100)
        {
            /* Start a transaction */
            rdm_status = d_trbegin("DeleteRecord", task);
            if (rdm_status == S_OKAY)
            {
                /* Lock the REC01 record */
                rdm_status = d_reclock(REC01, "w", task, CURR_DB);
                if (rdm_status == S_OKAY)
                {
                    /* Navigate to the smallest record */
                    rdm_status = d_keyfrst(REC01_FIELD01, task, CURR_DB);
                    if(rdm_status == S_OKAY)
                    {
                        /* Delete the record instance */
                        rdm_status = d_delete(task, CURR_DB);
                
                        /* Commit or rollback the tranaction */
                        if (rdm_status == S_OKAY)
                            d_trend(task);
                        else
                            d_trabort(task);
                    }
                }
            }
        }
        else
            sleep(1);
    }
}

static void Reader(void)
{
    DB_TASK   *task;
    int32_t    rdm_status;
    uint64_t   rec_count = 0;
    int32_t    field01_min;
    int32_t    field01_max;
    TFS_HANDLE hTFS;

    printf("Reader process\n");

    /* Initialize the application to use the external TFS (RPC) that we 
     * started
     */
    d_tfsinitEx(NULL, NULL, &hTFS, TFS_TYPE_RPC);

    /* Create a task for this thread */
    rdm_status = d_opentask(&task);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Error open DB_TASK in ReaderThreadFcn - %s: %s\n", d_errorname(rdm_status), d_errorstr(rdm_status));

        exit(EXIT_FAILURE);
    }

    /* Open the database in shared mode */
    rdm_status = d_open("core17_db", "s", task);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Error opening database in ReaderThreadFcn - %s: %s\n", d_errorname(rdm_status), d_errorstr(rdm_status));

        d_closetask(task);
        exit(EXIT_FAILURE);
    }

    printf("\n");
    printf("Total Records   Min Val  Max Val\n");
    printf("-------------   -------  -------\n");

    /* Loop until we get an error or are told to quit */
    while (rdm_status == S_OKAY)
    {
        /* Wait one second */
        sleep(1);

        /* Get information about the REC01 records */
        rdm_status = d_trrobegin(task, CURR_DB, NULL);
        if(rdm_status == S_OKAY)
        {
            d_rectot(REC01, &rec_count, task, CURR_DB);
            d_keyfrst(REC01_FIELD01, task, CURR_DB);
            d_crread(REC01_FIELD01, &field01_min, task, CURR_DB);
            d_keylast(REC01_FIELD01, task, CURR_DB);
            d_crread(REC01_FIELD01, &field01_max, task, CURR_DB);
            d_trroend(task, CURR_DB);
        }
        printf("       %6llu    %6d   %6d\n", rec_count, field01_min, field01_max);
    }
}

/* Start an external TFS Process */
/* Start an external TFS Process */
static pid_t start_tfs_process(void)
{
    pid_t pid = fork();
    if (pid < 0)
    {
        fprintf(stderr, "Error forking (%s)\n", strerror(errno));
        exit(EXIT_FAILURE);
    }

    if (pid == 0)
    {
        int32_t     stat;
        const char *tfserver_main_argv[] = {
            "tfserver"
        };

        (void) psp_init();

        stat = tfserver_main(1, tfserver_main_argv);
        if (stat != S_OKAY)
        {
            fprintf(stderr, "Error starting tfserver (%d)\n", stat);
            psp_term();
            exit(EXIT_FAILURE);
        }
    }

    return pid;
}

/* Stop a process */
static void stop_process(
    pid_t pid)
{
    int ret;
    int stat;

    /* Kill the tfserver process */
    ret = kill(pid, SIGINT);
    if (ret == -1)
    {
        fprintf(stderr, "Error killing tfserver (%s)\n", strerror(errno));
        exit(EXIT_FAILURE);
    }

    /* Wait for the tfserver process to terminate */
    do
    {
        waitpid(pid, &stat, 0);
    } while (!WIFEXITED(stat) && !WIFSIGNALED(stat));
}

/* =======================================================================
*/
int32_t main(
    int32_t            argc,
    const char *const *argv)
{
    DB_TASK   *task;
    int32_t    rdm_status;
    pid_t      tfs_pi;
    pid_t      inserter_pi;
    pid_t      deleter_pi;
    pid_t      reader_pi;
    TFS_HANDLE hTFS;

    printf("Starting tfserver (and waiting 5 seconds for it to initialize)\n");

    /* Start the tfserver process to process 'remote' connections */
    tfs_pi = start_tfs_process();
    if (tfs_pi > 0)
    {
        /* Sleep to give the tfserver time to initialize and start listening */
        sleep(5);

        /* Initialize the application to use the external TFS that we started */
        d_tfsinitEx(NULL, NULL, &hTFS, TFS_TYPE_RPC);

        /* create the context for initializing the database */
        d_opentask(&task);

        /* Open the database for exclusive access, this will create the database
         * dictionary file (core17_db.dbd) if it does not already exist.
         */
        rdm_status = d_open_ptr("core17_db", "x", &core17_db_dbd, sizeof(core17_db_dbd), task);
        if (rdm_status != S_OKAY)
        {
            fprintf(stderr, "Exclusive open failed - %s: %s\n", d_errorname(rdm_status), d_errorstr(rdm_status) );

            d_closetask(task);
            return EXIT_FAILURE;
        }

        /* Initialize the databases.  Initializing will create the data files if 
         * they do not exist.  If the the data files do already exist they will be
         * removed and new empty files will be created.
         */
        d_initialize(task, CURR_DB);

        /* Close the database in the main thread */
        d_close(task);    

        /* Start the insert, delete, and read processes */
        printf("\nStarting processes...\n");
        printf("Press <Enter> when you are ready to terminate them . . .\n");

        inserter_pi = fork();
        if (inserter_pi == 0)
            Inserter();
        else if (inserter_pi > 0)
        {
            deleter_pi = fork();
            if (deleter_pi == 0)
                Deleter();
            else if (deleter_pi > 0)
            {
                reader_pi = fork();
                if (reader_pi == 0)
                    Reader();
                else if (reader_pi > 0)
                {
                    /* Wait for user to hit <Enter> */
                    getchar();
    
                    d_closetask(task);
                    d_tfsterm(hTFS);

                    stop_process(reader_pi);
                    printf("Reader process finished\n");
                }

                stop_process(deleter_pi);
                printf("Deleter process finished\n");
            }

            stop_process(inserter_pi);
            printf("Inserter process finished\n");
        }

        /* Kill the tfserver process */
        stop_process(tfs_pi);
    }

    return EXIT_SUCCESS;
}
